from . import errors
from .base import BaseObject
from .collection import MontyCollection
from .types import encode_, string_types


INVALID_CHARS = ("$", "\0", "\x00")


class MontyDatabase(BaseObject):
    def __init__(self, client, name, codec_options=None, write_concern=None):
        super(MontyDatabase, self).__init__(
            codec_options or client.codec_options, write_concern or client.write_concern
        )

        self._client = client
        self._name = name
        self._components = (self,)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._client == other.client and self._name == other.name

        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "MontyDatabase({!r}, {!r})".format(self._client, self._name)

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(
                "MontyDatabase has no attribute {0!r}. To access the {0}"
                " collection, use database[{0!r}].".format(name)
            )
        return self.get_collection(name)

    def __getitem__(self, key):
        return self.get_collection(key)

    @property
    def name(self):
        """
        Return database's name.
        """
        return self._name

    @property
    def client(self):
        return self._client

    def collection_names(self):
        return self.client._storage.collection_list(self)

    list_collection_names = collection_names

    def create_collection(self, name, codec_options=None, write_concern=None, **kwargs):
        """
        Create a collection or raise an error if it already exists.
        """
        if self.client._storage.collection_exists(self, name):
            error_msg = "collection {} already exists".format(encode_(name))

            raise errors.CollectionInvalid(error_msg)

        collection = self.get_collection(name, codec_options, write_concern, **kwargs)
        self.client._storage.collection_create(self, name)

        return collection

    def drop_collection(self, name_or_collection):
        if isinstance(name_or_collection, MontyCollection):
            name = name_or_collection.name
        elif not isinstance(name_or_collection, string_types):
            raise TypeError("name_or_collection must be an instance of basestring")
        else:
            name = name_or_collection

        self.client._storage.collection_drop(self, name)

    def get_collection(self, name, codec_options=None, write_concern=None, **kwargs):
        """
        Return a collection if valid, otherwise throw an error.
        """
        if not name:
            raise errors.OperationFailure("Collection name may not be blank")

        for c in INVALID_CHARS:
            if c in name:
                raise errors.OperationFailure(
                    "Collection name contains special characters: {}".format(name)
                )

        if name.startswith("system."):
            raise errors.OperationFailure(
                "Invalid prefix in collection name: {}".format(name)
            )

        return MontyCollection(
            self, name, False, codec_options, write_concern, **kwargs
        )



import warnings
from copy import deepcopy

from .base import (
    BaseObject,
    validate_is_mapping,
    validate_ok_for_update,
    validate_ok_for_replace,
    validate_list_or_none,
    validate_boolean,
)

from .cursor import MontyCursor
from .engine.field_walker import FieldWalker
from .engine.weighted import Weighted
from .engine.queries import QueryFilter
from .engine.update import Updator
from .types import (
    abc,
    bson,
    string_types,
    is_duckument_type,
    Counter,
    on_err_close,
)
from .storage import StorageDuplicateKeyError
from .errors import (
    DuplicateKeyError,
    BulkWriteError,
    WriteError,
)

from .results import (
    DeleteResult,
    InsertOneResult,
    InsertManyResult,
    UpdateResult,
)


NotImplementeds = {
    "aggregate",
    "aggregate_raw_batches",
    "with_options",
    "bulk_write",
    "watch",
    "find_raw_batches",
    "find_one_and_delete",
    "find_one_and_replace",
    "find_one_and_update",
    "create_index",
    "create_indexes",
    "drop_index",
    "drop_indexes",
    "reindex",
    "list_indexes",
    "index_information",
    "rename",
    "options",
    "map_reduce",
    "inline_map_reduce",
    "parallel_scan",
}


class MontyCollection(BaseObject):
    def __init__(
        self,
        database,
        name,
        create=False,
        codec_options=None,
        write_concern=None,
        **kwargs
    ):
        """ """
        super(MontyCollection, self).__init__(
            codec_options or database.codec_options,
            write_concern or database.write_concern,
        )

        self._storage = database.client._storage

        self._database = database
        self._name = name
        self._components = (database, self)

    def __repr__(self):
        return "MontyCollection({!r}, {!r})".format(self._database, self._name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._database == other.database and self._name == other.name
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __getattr__(self, name):
        if name in NotImplementeds:
            raise NotImplementedError(
                "'MontyCollection.%s' is NOT implemented !" % name
            )
        if name.startswith("_"):
            full_name = ".".join((self._name, name))
            raise AttributeError(
                "MontyCollection has no attribute {0!r}. To access the {1}"
                " collection, use database[{1!r}].".format(name, full_name)
            )
        return self.__getitem__(name)

    def __getitem__(self, key):
        return self._database.get_collection(".".join((self._name, key)))

    @property
    def full_name(self):
        """ """
        return u".".join((self._database.name, self._name))

    @property
    def name(self):
        """ """
        return self._name

    @property
    def database(self):
        """ """
        return self._database

    def insert_one(self, document, bypass_document_validation=False, *args, **kwargs):
        """ """
        if bypass_document_validation:
            pass

        if "_id" not in document:
            document["_id"] = bson.ObjectId()

        try:
            result = self._storage.write_one(self, document)
        except StorageDuplicateKeyError:
            message = (
                "E11000 duplicate key error collection: %s index: "
                '_id_ dup key: { : "%s" }' % (self.full_name, str(document["_id"]))
            )
            details = {"index": 0, "code": 11000, "errmsg": message}
            raise DuplicateKeyError(message, code=11000, details=details)

        return InsertOneResult(result)

    def insert_many(
        self, documents, ordered=True, bypass_document_validation=False, *args, **kwargs
    ):
        """ """
        if not isinstance(documents, abc.Iterable) or not documents:
            raise TypeError("documents must be a non-empty list")

        if bypass_document_validation:
            pass

        def set_id(doc):
            if "_id" not in doc:
                doc["_id"] = bson.ObjectId()
            # Keep _id in track for error message
            return doc["_id"]

        counter = Counter(iter(documents), job_on_each=set_id)

        try:
            result = self._storage.write_many(self, counter, ordered)
        except StorageDuplicateKeyError:
            message = (
                "E11000 duplicate key error collection: %s index: "
                '_id_ dup key: { : "%s" }' % (self.full_name, str(counter.data))
            )
            index = counter.count - 1
            result = {
                "writeErrors": [
                    {
                        "index": index,
                        "code": 11000,
                        "errmsg": message,
                        "op": documents[index],
                    }
                ],
                "writeConcernErrors": [],
                "nInserted": index,
                "nUpserted": 0,
                "nMatched": 0,
                "nModified": 0,
                "nRemoved": 0,
                "upserted": [],
            }
            raise BulkWriteError(result)

        return InsertManyResult(result)

    def replace_one(
        self,
        filter,
        replacement,
        upsert=False,
        bypass_document_validation=False,
        *args,
        **kwargs
    ):
        """ """
        validate_is_mapping("filter", filter)
        validate_ok_for_replace(replacement)
        validate_boolean("upsert", upsert)

        if bypass_document_validation:
            pass

        raw_result = {"n": 0, "nModified": 0}
        # updator = Updator(replacement)
        try:
            fw = next(self._internal_scan_query(filter))
        except StopIteration:
            if upsert:
                if "_id" not in replacement:
                    replacement["_id"] = bson.ObjectId()
                raw_result["upserted"] = replacement["_id"]
                raw_result["n"] = 1
                self._storage.write_one(self, replacement, check_keys=False)
        else:
            raw_result["n"] = 1
            if fw.doc != replacement:
                replacement["_id"] = fw.doc["_id"]
                self._storage.update_one(self, replacement)
                raw_result["nModified"] = 1

        return UpdateResult(raw_result)

    def _internal_scan_query(self, query_spec):
        """An internal document generator for update"""
        queryfilter = QueryFilter(query_spec)
        documents = self._storage.query(MontyCursor(self), 0)
        first_matched = None
        for doc in documents:
            if queryfilter(doc):
                first_matched = queryfilter.fieldwalker
                break

        if first_matched:
            yield first_matched  # for try statement to test update or insert
            yield first_matched  # start update, yield again
            # continue iter documents(generator)
            for doc in documents:
                if queryfilter(doc):
                    yield queryfilter.fieldwalker

    def _internal_upsert(self, query_spec, updator, raw_result):
        """Internal document upsert"""
        doc_cls = self._database.codec_options.document_class

        def _remove_dollar_key(doc):
            if is_duckument_type(doc):
                new_doc = doc_cls()
                fields = list(doc.keys())
                for field in fields:
                    if field[:1] == "$" or "." in field:
                        continue
                    new_doc[field] = _remove_dollar_key(doc[field])
                return new_doc
            else:
                return doc

        document = _remove_dollar_key(deepcopy(query_spec))
        if "_id" not in document:
            document["_id"] = bson.ObjectId()
        raw_result["upserted"] = document["_id"]
        raw_result["n"] = 1

        fieldwalker = FieldWalker(document)
        updator(fieldwalker, do_insert=True)
        self._storage.write_one(self, fieldwalker.doc)

    def _no_id_update(self, updator, filter=None):
        id_operator = updator.operations.get("_id")
        doc_id = (filter or {}).get("_id")
        if id_operator and id_operator._keep() != doc_id:
            msg = (
                "Performing an update on the path '_id' would "
                "modify the immutable field '_id'"
            )
            raise WriteError(msg, code=66)

    def update_one(
        self,
        filter,
        update,
        upsert=False,
        bypass_document_validation=False,
        array_filters=None,
        *args,
        **kwargs
    ):
        """ """
        validate_is_mapping("filter", filter)
        validate_ok_for_update(update)
        validate_list_or_none("array_filters", array_filters)
        validate_boolean("upsert", upsert)

        if bypass_document_validation:
            pass

        raw_result = {"n": 0, "nModified": 0}
        updator = Updator(update, array_filters)
        self._no_id_update(updator, filter)
        try:
            fw = next(self._internal_scan_query(filter))
        except StopIteration:
            if upsert:
                self._internal_upsert(filter, updator, raw_result)
        else:
            self._no_id_update(updator)

            raw_result["n"] = 1
            if updator(fw):
                self._storage.update_one(self, fw.doc)
                raw_result["nModified"] = 1

        return UpdateResult(raw_result)

    def update_many(
        self,
        filter,
        update,
        upsert=False,
        bypass_document_validation=False,
        array_filters=None,
        *args,
        **kwargs
    ):
        """ """
        validate_is_mapping("filter", filter)
        validate_ok_for_update(update)
        validate_list_or_none("array_filters", array_filters)
        validate_boolean("upsert", upsert)

        if bypass_document_validation:
            pass

        raw_result = {"n": 0, "nModified": 0}
        updator = Updator(update, array_filters)
        scanner = self._internal_scan_query(filter)
        self._no_id_update(updator, filter)
        try:
            next(scanner)
        except StopIteration:
            if upsert:
                self._internal_upsert(filter, updator, raw_result)
        else:
            self._no_id_update(updator)

            @on_err_close(scanner)
            def update_docs():
                n, m = 0, 0
                for fieldwalker in scanner:
                    n += 1
                    if updator(fieldwalker):
                        m += 1
                        yield fieldwalker.doc
                raw_result["n"] = n
                raw_result["nModified"] = m

            self._storage.update_many(self, update_docs())

        return UpdateResult(raw_result)

    def delete_one(self, filter):
        raw_result = {"n": 0}

        queryfilter = QueryFilter(filter)
        storage = self._storage
        documents = storage.query(MontyCursor(self), 0)

        for doc in documents:
            if queryfilter(doc):
                storage.delete_one(self, doc["_id"])
                raw_result["n"] = 1
                break

        return DeleteResult(raw_result)

    def delete_many(self, filter):
        raw_result = {"n": 0}

        queryfilter = QueryFilter(filter)
        storage = self._storage
        documents = storage.query(MontyCursor(self), 0)

        doc_ids = set()
        for doc in documents:
            if queryfilter(doc):
                doc_ids.add(doc["_id"])
                raw_result["n"] += 1

        storage.delete_many(self, doc_ids)

        return DeleteResult(raw_result)

    def find(self, *args, **kwargs):
        # return a cursor
        return MontyCursor(self, *args, **kwargs)

    def find_one(self, filter=None, *args, **kwargs):
        """ """
        if filter is not None and not isinstance(filter, abc.Mapping):
            filter = {"_id": filter}

        cursor = self.find(filter, *args, **kwargs)
        for result in cursor.limit(-1):
            return result
        return None

    def count(self, filter=None, **kwargs):
        warnings.warn(
            "count is deprecated. Use Collection.count_documents instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.count_documents(filter, **kwargs)

    def count_documents(self, filter, **kwargs):
        cursor = MontyCursor(self, filter=filter, **kwargs)
        return len(list(cursor))

    def distinct(self, key, filter=None, **kwargs):
        """ """
        if not isinstance(key, string_types):
            raise TypeError(
                "key must be an instance of %s" % (string_types.__name__,)
            )

        result = list()

        def get_value(doc):
            fieldwalker = FieldWalker(doc)
            fieldvalues = fieldwalker.go(key).get().value
            res = list()
            for v in fieldvalues.iter_flat():
                weighted = Weighted(v)
                if weighted not in result:
                    res.append(weighted)
            return res

        documents = self._storage.query(MontyCursor(self), 0)

        if filter:
            queryfilter = QueryFilter(filter)
            for doc in documents:
                if queryfilter(doc):
                    result += get_value(doc)
        else:
            for doc in documents:
                result += get_value(doc)

        return [weighted.value for weighted in sorted(result)]

    def drop(self):
        self._database.drop_collection(self._name)

    def save(self, to_save, *args, **kwargs):
        # DEPRECATED
        if "_id" in to_save:
            self.replace_one(
                {"_id": to_save["_id"]}, to_save, upsert=True, *args, **kwargs
            )
        else:
            self.insert_one(to_save, *args, **kwargs)




import warnings
from .bson import init, bson_used  # noqa
from .bson import *  # noqa


warnings.warn(
    "montydb.types.bson_ is deprecated; import montydb.types.bson instead",
    DeprecationWarning
)



"""Managing all data types and mocks
"""

from .compat import (
    PY3,
    ConfigParser,
    abc,
    zip_longest,
    integer_types,
    string_types,
    text_type,
    iteritems,
    unicode_,
    encode_,
    decode_,
)
from .helper import (
    RE_PATTERN_TYPE,
    to_bytes,
    compare_documents,
    is_duckument_type,
    is_integer_type,
    is_numeric_type,
    is_pattern_type,
    re_int_flag_to_str,
    re_str_flags_to_int,

    Counter,
    keep,
    on_err_close,
)
from .bson import init as init_bson


__all__ = [
    "PY3",
    "ConfigParser",
    "abc",
    "zip_longest",
    "integer_types",
    "string_types",
    "text_type",
    "iteritems",
    "unicode_",
    "encode_",
    "decode_",

    "RE_PATTERN_TYPE",
    "to_bytes",
    "compare_documents",
    "is_duckument_type",
    "is_integer_type",
    "is_numeric_type",
    "is_pattern_type",
    "re_int_flag_to_str",
    "re_str_flags_to_int",

    "Counter",
    "keep",
    "on_err_close",

    "init_bson",
]



from __future__ import absolute_import
import re
import sys
import platform
from .compat import integer_types, zip_longest
from . import bson

RE_PATTERN_TYPE = type(re.compile(""))


def to_bytes(s):
    return s.encode()


def _compare_doc_in_strict_order(a, b):
    if len(a) == len(b):
        if all(ak == bk for ak, bk in zip_longest(a.keys(), b.keys())):
            for av, bv in zip_longest(a.values(), b.values()):
                if (
                    is_duckument_type(av)
                    and is_duckument_type(bv)
                    and not _compare_doc_in_strict_order(av, bv)
                ) or av != bv:
                    return False
            return True
    return False


DictKeyOrderMutable = (
    # CPython >= 3.5
    sys.version_info >= (3, 5)
    # all PyPy
    or platform.python_implementation() == "PyPy"
)


def compare_documents(this, that):
    """Document key order matters

    In PY 3.6, `dict` has order-preserving, so does PyPy and PyPy3, and in
    PY 3.5, `dict` is random iteration ordered.

    Which means, same key-value pairs could have different orders in runtime,
    and two dictionary with same values but different key ordered are still
    equal in comparison, but not equal in MongoDB.

    Therefore, we should compare them by length, keys, values separately.
    """
    if not DictKeyOrderMutable:
        # use simple `equal` if the order is constant
        return this == that
    else:
        return _compare_doc_in_strict_order(this, that)


def is_duckument_type(obj):
    """Internal mapping type checker

    Instead of using `isinstance(obj, MutableMapping)`, duck type checking
    is much cheaper and work on most common use cases.

    If an object has these attritubes, is a document:
        `__len__`, `keys`, `values`

    """
    doc_attrs = ("__len__", "keys", "values")
    return all(hasattr(obj, attr) for attr in doc_attrs)


def is_numeric_type(obj):
    if isinstance(obj, bool):
        return False
    return isinstance(obj, (integer_types, float, bson.Int64, bson.Decimal128))


def is_integer_type(obj):
    if isinstance(obj, bool):
        return False
    return isinstance(obj, (integer_types, bson.Int64))


def is_pattern_type(obj):
    return isinstance(obj, RE_PATTERN_TYPE)


def re_int_flag_to_str(int_flags):
    """Ripped from bson.json_util"""
    flags = ""
    if int_flags & re.IGNORECASE:
        flags += "i"
    if int_flags & re.LOCALE:
        flags += "l"
    if int_flags & re.MULTILINE:
        flags += "m"
    if int_flags & re.DOTALL:
        flags += "s"
    if int_flags & re.UNICODE:
        flags += "u"
    if int_flags & re.VERBOSE:
        flags += "x"
    return flags


def re_str_flags_to_int(str_flags):
    flags = 0
    if "i" in str_flags:
        flags |= re.IGNORECASE
    if "l" in str_flags:
        flags |= re.LOCALE
    if "m" in str_flags:
        flags |= re.MULTILINE
    if "s" in str_flags:
        flags |= re.DOTALL
    if "u" in str_flags:
        flags |= re.UNICODE
    if "x" in str_flags:
        flags |= re.VERBOSE

    return flags


def __add_attrib(deco):
    """Decorator helper"""

    def meta_decorator(value):
        def add_attrib(func):
            func._keep = lambda: value
            return func

        return add_attrib

    return meta_decorator


@__add_attrib
def keep(query):
    """A decorator that preserve operation query for operator"""

    def _(func):
        return query

    return _


class Counter(object):
    """An iterator that could trace the progress of iteration

    Args:
        iterable (Iterable): An iterable object.
        job_on_each (func): The function to process each item in the iterable
                            during the iteration. `function(item)`
                            The return value of the function will be stored in
                            `Counter.data` attribute.

    Attributes:
        count (int): An int that indicate the progress of iteration.
        data: The value returned from the `job_on_each` function.

    """

    def __init__(self, iterable, job_on_each=None):
        self._iterable = iterable
        self._job = job_on_each or (lambda i: None)
        self.count = 0
        self.data = None

    def __next__(self):
        item = next(self._iterable)
        self.data = self._job(item)
        self.count += 1
        return item

    next = __next__

    def __iter__(self):
        return self


def on_err_close(generator):
    """Decorator, to close `generator` passed into the function on error

    Ensure the `generator` which holds the storage database connection to
    close that connection right on error occurred and then raise the error.

    In some tests, one may leave the connection open and cause next test
    case fail on startup or teardown, e.g. drop collection that created by
    previous test. The reason why that previous connection remains open,
    could be the `generator` which holds the connection context was not
    being garbage collected (`__exit__` was not triggered) while entring
    the next test. So we need this at least for now, until better solution
    or better test cases came up.

    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                for i in func(*args, **kwargs):
                    yield i
            except Exception:
                generator.close()
                raise

        return wrapper

    return decorator



from __future__ import absolute_import
import sys

bson_used = None


SON = None
BSON = None
ObjectId = None
Timestamp = None
MinKey = None
MaxKey = None
Int64 = None
Decimal128 = None
Binary = None
Regex = None
Code = None
RawBSONDocument = None
CodecOptions = None

decimal128_NaN = None
decimal128_INF = None
decimal128_NaN_ls = None

BSONError = None
InvalidId = None
InvalidDocument = None

id_encode = None
document_encode = None
document_decode = None
json_loads = None
json_dumps = None
parse_codec_options = None


def init(use_bson=None):
    from .. import errors

    self = sys.modules[__name__]
    if self.bson_used is not None:
        return

    # Init

    if use_bson is None:
        try:
            import bson  # noqa: F401
        except ImportError:
            use_bson = False
        else:
            use_bson = True

    if use_bson:
        from ._bson import BSON_
        bson_ = BSON_()
    else:
        from ._nobson import NoBSON
        bson_ = NoBSON()

    self.bson_used = bson_.bson_used
    for name in __all__:
        setattr(self, name, getattr(bson_, name))

    errors.init_bson_err()


__all__ = [
    "SON",
    "BSON",
    "ObjectId",
    "Timestamp",
    "MinKey",
    "MaxKey",
    "Int64",
    "Decimal128",
    "Binary",
    "Regex",
    "Code",
    "RawBSONDocument",
    "CodecOptions",

    "decimal128_NaN",
    "decimal128_INF",
    "decimal128_NaN_ls",

    "BSONError",
    "InvalidId",
    "InvalidDocument",

    "id_encode",
    "document_encode",
    "document_decode",
    "json_loads",
    "json_dumps",
    "parse_codec_options",
]
